
from infi import unittest
import __builtin__
import mock

#pylint: disable-all

MOCK_OUTPUT = [
               {'show multipaths topology': "create: 36000402001f45eb566e79f6d00000000 dm-0 NEXSAN,SATABoy2\nsize=3.7G features='0' hwhandler='0' wp=rw\n`-+- policy='round-robin 0' prio=1 status=active\n  `- 4:0:0:10 sdc 8:32 active ready  running\ncreate: 36000402001f45eb56424ca6800000000 dm-1 NEXSAN,SATABoy2\nsize=466G features='0' hwhandler='0' wp=rw\n`-+- policy='round-robin 0' prio=0 status=active\n  `- 4:0:0:0  sdb 8:16 failed faulty running\ncreate: 36000402001f45eb566e79fb700000000 dm-2 NEXSAN,SATABoy2\nsize=5.6G features='0' hwhandler='0' wp=rw\n`-+- policy='round-robin 0' prio=1 status=active\n  `- 4:0:0:11 sdd 8:48 active ready  running",
                'show paths': 'hcil     dev dev_t pri dm_st  chk_st dev_st  next_check     \n2:0:0:0  sda 8:0   1   undef  ready  running orphan         \n4:0:0:10 sdc 8:32  1   active ready  running XXXX...... 9/20\n4:0:0:0  sdb 8:16  1   failed faulty running X......... 2/20\n4:0:0:11 sdd 8:48  1   active ready  running XXXX...... 8/20'
                },
               {'show multipaths topology': 'create: 36000402001f45eb56424ca6800000000dm-0  NEXSAN  ,SATABoy2      \n[size=466G][features=0       ][hwhandler=0        ]\n\\_ round-robin 0 [prio=0][enabled]\n \\_ 5:0:0:0  sdb 8:16  [active][ready]\ncreate: 36000402001f45eb566e79f6d00000000dm-1  NEXSAN  ,SATABoy2      \n[size=3.7G][features=0       ][hwhandler=0        ]\n\\_ round-robin 0 [prio=0][enabled]\n \\_ 5:0:0:10 sdc 8:32  [active][ready]\ncreate: 36000402001f45eb566e79fb700000000dm-2  NEXSAN  ,SATABoy2      \n[size=5.6G][features=0       ][hwhandler=0        ]\n\\_ round-robin 0 [prio=0][enabled]\n \\_ 5:0:0:11 sdd 8:48  [active][ready]',
                 'show paths': 'hcil     dev dev_t pri dm_st   chk_st  next_check      \n2:0:0:0  sda 8:0   1   [undef] [undef] [orphan]        \n5:0:0:0  sdb 8:16  1   [active][ready] .......... 1/20 \n5:0:0:10 sdc 8:32  1   [active][ready] XXXXXXX... 15/20\n5:0:0:11 sdd 8:48  1   [active][ready] XXXXXXXX.. 17/20'
                },
               {'show multipaths topology': 'create: 36000402001f45eb56424ca6800000000 dm-0  NEXSAN,SATABoy2\n[size=466G][features=0       ][hwhandler=0        ][rw        ]\n\\_ round-robin 0 [prio=0][enabled]\n \\_ 2:0:0:0  sdb 8:16  [active][ready]\ncreate: 36000402001f45eb566e79f6d00000000 dm-1  NEXSAN,SATABoy2\n[size=3.7G][features=0       ][hwhandler=0        ][rw        ]\n\\_ round-robin 0 [prio=0][active]\n \\_ 2:0:0:10 sdc 8:32  [active][ready]\ncreate: 36000402001f45eb566e79fb700000000 dm-2  NEXSAN,SATABoy2\n[size=5.6G][features=0       ][hwhandler=0        ][rw        ]\n\\_ round-robin 0 [prio=0][enabled]\n \\_ 2:0:0:11 sdd 8:48  [active][ready]',
                'show paths': 'hcil     dev dev_t pri dm_st   chk_st  next_check     \n0:0:0:0  sda 8:0   1   [undef] [ready] [orphan]       \n2:0:0:0  sdb 8:16  1   [active][ready] XXXX...... 9/20\n2:0:0:10 sdc 8:32  1   [active][ready] X......... 2/20\n2:0:0:11 sdd 8:48  1   [active][ready] XXXX...... 9/20'
                },
               {'show multipaths topology': "mpatha (35742b0f006800000) dm-0 INFINID,Infinidat A01\nsize=10G features='0' hwhandler='0' wp=rw\n|-+- policy='round-robin 0' prio=1 status=active\n| `- 3:0:0:1 sdb 8:16 active ready running\n`-+- policy='round-robin 0' prio=1 status=enabled\n  `- 2:0:0:1 sdc 8:32 active ready running\nmpathb (36000402001f45eb565889a4b00000000) dm-2 NEXSAN,SATABoy2\nsize=47G features='0' hwhandler='0' wp=rw\n|-+- policy='round-robin 0' prio=1 status=active\n| `- 2:0:1:0 sdd 8:48 active ready running\n`-+- policy='round-robin 0' prio=1 status=enabled\n  `- 3:0:1:1 sdf 8:80 active ready running\nmpathc (36000402001f45eb565a24ae100000000) dm-3 NEXSAN,SATABoy2\nsize=466G features='0' hwhandler='0' wp=rw\n`-+- policy='round-robin 0' prio=1 status=active\n  `- 3:0:1:0 sde 8:64 active ready running",
                'show paths': 'hcil    dev dev_t pri dm_st  chk_st dev_st  next_check      \n4:0:0:0 sda 8:0   1   undef  ready  running orphan          \n3:0:0:1 sdb 8:16  1   active ready  running XXXXXXX... 14/20\n2:0:0:1 sdc 8:32  1   active ready  running XXXXXXXX.. 16/20\n2:0:1:0 sdd 8:48  1   active ready  running XXXXXXXX.. 16/20\n3:0:1:0 sde 8:64  1   active ready  running XXXXXXX... 14/20\n3:0:1:1 sdf 8:80  1   active ready  running XXXXXX.... 13/20'
                },
               {'show multipaths topology': "mpatha (35742b0f006800000) dm-0 INFINID,Infinidat A01\nsize=10G features='0' hwhandler='0' wp=rw\n|-+- policy='round-robin 0' prio=1 status=active\n| `- 3:0:0:1 sdb 8:16 active ready running\n`-+- policy='round-robin 0' prio=1 status=enabled\n  `- 2:0:0:1 sdc 8:32 active ready running\nmpathb (36000402001f45eb565889a4b00000000) dm-2 NEXSAN,SATABoy2\nsize=47G features='0' hwhandler='0' wp=rw\n`-+- policy='round-robin 0' prio=1 status=active\n  |- 2:0:1:0 sdd 8:48 active ready running\n  `- 3:0:1:1 sdf 8:80 active ready running\nmpathc (36000402001f45eb565a24ae100000000) dm-3 NEXSAN,SATABoy2\nsize=466G features='0' hwhandler='0' wp=rw\n`-+- policy='round-robin 0' prio=1 status=active\n  `- 3:0:1:0 sde 8:64 active ready running",
                'show paths': 'hcil    dev dev_t pri dm_st  chk_st dev_st  next_check      \n4:0:0:0 sda 8:0   1   undef  ready  running orphan          \n3:0:0:1 sdb 8:16  1   active ready  running XXXXXXXX.. 16/20\n2:0:0:1 sdc 8:32  1   active ready  running XXXXXXX... 15/20\n2:0:1:0 sdd 8:48  1   active ready  running XXXX...... 9/20 \n3:0:1:0 sde 8:64  1   active ready  running XXXXXXXXXX 20/20\n3:0:1:1 sdf 8:80  1   active ready  running XXXXX..... 10/20'
                }
               ]

from . import parse_paths_table, parse_multipaths_topology, strip_ansi_colors
from . import get_list_of_multipath_devices_from_multipathd_output
from ..dtypes import MultipathDevice, Path, PathGroup

class PathTableTestCase(unittest.TestCase):

    def _assert_lists(self, a, b):
        a.sort()
        b.sort()
        self.assertEquals(a, b)

    def _validate_example(self, matches):
        self.assertEqual(len(matches), 4)
        for hctl in [match['hctl'] for match in matches]:
            self.assertTrue(len(hctl.split(':')), 4)
            self.assertTrue([item.isdigit() for item in hctl.split(':')],
                            [True, ] * 4)
        self._assert_lists([match['dev'] for match in matches],
                         ['sda', 'sdc', 'sdb', 'sdd'])
        self._assert_lists([match['dev_t'] for match in matches],
                         ['8:0', '8:32', '8:16', '8:48'])
        for item in [match['dm_st'] for match in matches]:
            self.assertIn(item, ['active', 'undef', 'failed'])
        for item in [match['chk_st'] for match in matches]:
            self.assertIn(item, ['active', 'undef', 'failed', 'ready', 'faulty'])

    def test_parser__1(self):
        subject = MOCK_OUTPUT[0]['show paths']
        matches = [match for match in parse_paths_table(subject)]
        self._validate_example(matches)

    def test_parser__2(self):
        subject = MOCK_OUTPUT[1]['show paths']
        matches = [match for match in parse_paths_table(subject)]
        self.assertEqual(len(matches), 4)
        self._validate_example(matches)

    def test_parser__3(self):
        subject = MOCK_OUTPUT[2]['show paths']
        matches = [match for match in parse_paths_table(subject)]
        self.assertEqual(len(matches), 4)
        self._validate_example(matches)

    def test_parser__4(self):
        subject = MOCK_OUTPUT[3]['show paths']
        matches = [match for match in parse_paths_table(subject)]
        self.assertEqual(len(matches), 6)

    def test_parser__5(self):
        subject = MOCK_OUTPUT[4]['show paths']
        matches = [match for match in parse_paths_table(subject)]
        self.assertEqual(len(matches), 6)

class MultipathsTopologyTestCase(unittest.TestCase):
    def _validate_example(self, matches):
        self.assertEqual(len(matches), 3)
        self.assertEqual([len(match['path_groups']) for match in matches],
                         [1] * 3)
        self.assertEqual([len(match['path_groups'][0]['paths']) for match in matches],
                         [1] * 3)

    def test_parser__1(self):
        subject = MOCK_OUTPUT[0]['show multipaths topology']
        matches = [match for match in parse_multipaths_topology(subject)]
        self._validate_example(matches)

    def test_parser__2(self):
        subject = MOCK_OUTPUT[1]['show multipaths topology']
        matches = [match for match in parse_multipaths_topology(subject)]
        self.assertEqual(len(matches), 3)

    def test_parser__3(self):
        subject = MOCK_OUTPUT[2]['show multipaths topology']
        matches = [match for match in parse_multipaths_topology(subject)]
        self.assertEqual(len(matches), 3)

    def test_parser__4(self):
        subject = MOCK_OUTPUT[3]['show multipaths topology']
        matches = [match for match in parse_multipaths_topology(subject)]
        self.assertEqual(len(matches), 3)

    def test_parser__5(self):
        subject = MOCK_OUTPUT[4]['show multipaths topology']
        matches = [match for match in parse_multipaths_topology(subject)]
        self.assertEqual(len(matches), 3)

class ModelTestCase(unittest.TestCase):
    def _get_devices_from_example_by_index(self, index):
        output = MOCK_OUTPUT[index]
        maps_topology = output['show multipaths topology']
        paths_table = output['show paths']
        with mock.patch.object(__builtin__, "open") as open:
            devices = get_list_of_multipath_devices_from_multipathd_output(maps_topology, paths_table)
        self.assertEqual(len(devices), 3)
        return devices

    def test_example__1(self):
        devices = self._get_devices_from_example_by_index(0)
        actual = devices[0]
        with mock.patch.object(__builtin__, "open") as open:
            expected = MultipathDevice('36000402001f45eb566e79f6d00000000',
                                       '36000402001f45eb566e79f6d00000000',
                                       'dm-0')
        expected.path_groups.append(PathGroup('active', '1'))
        expected.path_groups[0].paths.append(Path('sdc', 'sdc', '8:32', 'active', '1', '4:0:0:10'))
        self.assertEqual(actual, expected)

    def test_example__4(self):
        devices = self._get_devices_from_example_by_index(3)
        self.assertEqual(len(devices), 3)
        actual = devices[0]
        with mock.patch.object(__builtin__, "open") as open:
            expected = MultipathDevice('35742b0f006800000',
                                       'mpatha',
                                       'dm-0')
        expected.path_groups.append(PathGroup('active', '1'))
        expected.path_groups.append(PathGroup('enabled', '1'))
        expected.path_groups[0].paths.append(Path('sdb', 'sdb', '8:16', 'active', '1',
                                                  '3:0:0:1'))
        expected.path_groups[1].paths.append(Path('sdc', 'sdc', '8:32', 'active', '1',
                                                  '2:0:0:1'))
        self.assertEqual(actual, expected)
        self.assertEqual(actual.path_groups[0].paths[0].hctl, (3, 0, 0, 1))

    def test_example__5(self):
        devices = self._get_devices_from_example_by_index(4)
        self.assertEqual(len(devices), 3)
        actual = devices[1]
        with mock.patch.object(__builtin__, "open") as open:
            expected = MultipathDevice('36000402001f45eb565889a4b00000000',
                                       'mpathb',
                                       'dm-2')
        expected.path_groups.append(PathGroup('active', '1'))
        expected.path_groups[0].paths.append(Path('sdd', 'sdd', '8:48', 'active', '1', '2:0:1:0'))
        expected.path_groups[0].paths.append(Path('sdf', 'sdf', '8:80', 'active', '1', '3:0:1:1'))
        self.assertEqual(actual, expected)

class AnsiColorsTestCase(unittest.TestCase):
    def test_strip(self):
        subject = "\n".join(["\x1b[1mcreate: 36000402001f45eb566e79f6d00000000 dm-0 NEXSAN,SATABoy2\x1b[0m"] * 4)
        expected = "\n".join(["create: 36000402001f45eb566e79f6d00000000 dm-0 NEXSAN,SATABoy2"] * 4)
        actual = strip_ansi_colors(subject)
        self.assertEqual(actual, expected)

VERSION_OUTPUT = ["""multipath-tools v0.4.8 (08/02, 2007)
CLI commands reference:
 list|show paths
 list|show maps|multipaths
 list|show maps|multipaths status
 list|show maps|multipaths stats
 list|show maps|multipaths topology
 list|show topology""",

 """multipath-tools v0.4.9 (04/04, 2009)
CLI commands reference:
 list|show paths
 list|show paths format $format
 list|show status""",

 """fail
multipath-tools v0.4.7 (03/12, 2006)
CLI commands reference:
 list|show paths
 list|show maps|multipaths
 list|show maps|multipaths status
 list|show maps|multipaths stats"""]

class ParseVersionTestCase(unittest.TestCase):
    @unittest.parameters.iterate("output", VERSION_OUTPUT)
    def test_parse_version(self, output):
        from . import parse_multipath_tools_version
        actual = parse_multipath_tools_version(output)
        expected = ["0.4.8", "0.4.9", "0.4.7"]
        self.assertIn(actual, expected)




import unittest
import mock
import os

try:
    from gevent import socket
except ImportError:
    import socket

socket_module = socket
from contextlib import contextmanager

from ctypes import sizeof, c_size_t

#pylint: disable-all

TEST_MESSAGE_TO_SEND = 'reconfigure'
TEST_MESSAGE_SIZE_TO_SEND = '\x0b' + '\x00' * (sizeof(c_size_t) - 1)
TEST_MESSAGE_SIZE_TO_RECEIVE = '\x04' + '\x00' * (sizeof(c_size_t) - 1)
TEST_MESSAGE_TO_RECEIVE = 'ok\n\x00'

class ConnectionTestCase(unittest.TestCase):
    def _get_connection_object(self):
        raise unittest.SkipTest

    def setUp(self):
        self.connection = self._get_connection_object()

    def test_connect(self):
        self.connection.connect()

    def test_connect_and_disconnect(self):
        self.connection.connect()
        self.connection.disconnect()

    def test_send_and_receive(self):
        self.connection.connect()
        self.connection.send(TEST_MESSAGE_SIZE_TO_SEND)
        self.connection.send(TEST_MESSAGE_TO_SEND)
        self.assertEquals(self.connection.receive(len(TEST_MESSAGE_SIZE_TO_RECEIVE)),
                          TEST_MESSAGE_SIZE_TO_RECEIVE)
        self.assertEquals(self.connection.receive(len(TEST_MESSAGE_TO_RECEIVE)),
                          TEST_MESSAGE_TO_RECEIVE)
        self.connection.disconnect()

class UnixDomainSocketTestCase(ConnectionTestCase):
    def _get_connection_object(self):
        from . import  UnixDomainSocket
        return UnixDomainSocket()

    def setUp(self):
        from platform import system
        if system().lower() != 'linux':
            raise unittest.SkipTest
        super(UnixDomainSocketTestCase, self).setUp()

def send_mock(*args, **kwargs):
    message = kwargs.get('string', args[0])
    return len(message)

def recv_mock(items_to_return):
    items = [item for item in items_to_return]
    def side_effect(*args, **kwargs):
        return items.pop()
    return side_effect

class MockUnixDomainSocketTestCase(UnixDomainSocketTestCase):
    def setUp(self):
        with mock.patch("platform.system") as system:
            system.return_value = 'Linux'
            super(UnixDomainSocketTestCase, self).setUp()

    @contextmanager
    def mock_socket(self):
        with mock.patch("infi.multipathtools.connection.socket.socket") as socket:
            socket.return_value = mock.Mock()
            socket.return_value.send.side_effect = send_mock
            yield socket

    def test_connect(self):
        with self.mock_socket() as socket:
            super(MockUnixDomainSocketTestCase, self).test_connect()
        self.assertTrue(socket.called)
        self.assertEquals(socket.call_count, 1)
        self.assertEquals(socket.call_args[0], (socket_module.AF_UNIX, socket_module.SOCK_STREAM))

    def test_connect_and_disconnect(self):
        with self.mock_socket() as socket:
            super(MockUnixDomainSocketTestCase, self).test_connect_and_disconnect()
        self.assertTrue(socket.called)
        self.assertEquals(socket.call_count, 1)
        self.assertEquals(socket.call_args[0], (socket_module.AF_UNIX, socket_module.SOCK_STREAM))
        self.assertEquals(socket.return_value.close.call_count, 1)

    def test_send_and_receive(self):
        with self.mock_socket() as socket:
            _recv_mock = recv_mock([TEST_MESSAGE_TO_RECEIVE, TEST_MESSAGE_SIZE_TO_RECEIVE])
            socket.return_value.recv.side_effect = _recv_mock
            super(MockUnixDomainSocketTestCase, self).test_send_and_receive()
        self.assertEquals(socket.return_value.send.call_count, 2)

    def _test_send_and_receive_long_messages(self):
        from . import MessageLength
        message_to_send = 'abcd' * 2048
        self.connection.connect()
        instance = MessageLength()
        instance.length = len(message_to_send)
        self.connection.send(MessageLength.write_to_string(instance))
        self.connection.send(message_to_send)
        self.assertEquals(self.connection.receive(2048), 'a' * 2048)
        self.connection.disconnect()

    def test_send_and_receive_long_messages(self):
        raise unittest.SkipTest
        def send_mock(*args, **kwargs):
            message = kwargs.get('string', args[0])
            return 1 if len(message) == 1 else len(message) / 2

        def recv_mock(*args, **kwargs):
            expected_size = kwargs.get('bufsize', args[0])
            return 'a' * (1 if expected_size == 1 else expected_size / 2)

        with self.mock_socket() as socket:
            socket.return_value.send.side_effect = send_mock
            socket.return_value.recv.side_effect = recv_mock
            self._test_send_and_receive_long_messages()
        self.assertEquals(socket.return_value.send.call_count, 18)
        self.assertEquals(socket.return_value.recv.call_count, 23)

    def test_connect__init_raises_exception(self):
        from socket import error
        from ..errors import ConnectionError
        with self.mock_socket() as socket:
            socket.side_effect = error
            self.assertRaises(ConnectionError, self.connection.connect)

    def test_connect__connect_raises_exception(self):
        from socket import error
        from ..errors import ConnectionError
        with self.mock_socket() as socket:
            socket.return_value.connect.side_effect = error
            self.assertRaises(ConnectionError, self.connection.connect)

    def test_connect__settimeout_raises_exception(self):
        from socket import error
        from ..errors import ConnectionError
        with self.mock_socket() as socket:
            socket.return_value.settimeout.side_effect = error
            self.assertRaises(ConnectionError, self.connection.connect)

    def test_send__raises_error(self):
        from socket import error
        from ..errors import ConnectionError
        with self.mock_socket() as socket:
            socket.return_value.send.side_effect = error
            self.connection.connect()
            self.assertRaises(ConnectionError, self.connection.send, *['foo'])

    def test_send__raises_timeout(self):
        from socket import timeout
        from ..errors import TimeoutExpired
        with self.mock_socket() as socket:
            socket.return_value.send.side_effect = timeout
            self.connection.connect()
            self.assertRaises(TimeoutExpired, self.connection.send, *['foo'])

    def test_recv__raises_error(self):
        from socket import error
        from ..errors import ConnectionError
        with self.mock_socket() as socket:
            socket.return_value.recv.side_effect = error
            self.connection.connect()
            self.assertRaises(ConnectionError, self.connection.receive, *[1])

    def test_recv__raises_timeout(self):
        from socket import timeout
        from ..errors import TimeoutExpired
        with self.mock_socket() as socket:
            socket.return_value.recv.side_effect = timeout
            self.connection.connect()
            self.assertRaises(TimeoutExpired, self.connection.receive, *[1])

    def _get_random_socket_path(self):
        import string
        import random
        N = 10
        return '/tmp/' + ''.join(random.choice(string.ascii_uppercase + string.digits) for x in range(N))

    def _get_socket_filepath(self):
        filepath = None
        while filepath is None or os.path.exists(filepath):
            filepath = self._get_random_socket_path()
        return filepath

    def _setup_unix_domain_socket(self):
        filepath = self._get_socket_filepath()
        sock = socket.socket(socket_module.AF_UNIX, socket_module.SOCK_STREAM)
        sock.bind(filepath)
        sock.listen(1)
        sock.setblocking(0)
        _ = sock.accept
        self.addCleanup(lambda: sock.close())
        return sock, filepath

    def test_receive__no_premature_timeout(self):
        from . import UnixDomainSocket
        from ..errors import TimeoutExpired
        from time import time
        server_side_socket, socket_filepath = self._setup_unix_domain_socket()
        self.assertTrue(os.path.exists(socket_filepath))
        client_side_socket = UnixDomainSocket(address=socket_filepath, timeout=5)
        client_side_socket.connect()
        before = time()
        with self.assertRaises(TimeoutExpired):
            client_side_socket.receive()
        after = time()
        self.assertGreater(after-before, 3)

    def test_repr(self):
        from . import UnixDomainSocket
        string = repr(UnixDomainSocket())




#pylint: disable-msg=E1101

from os.path import sep, join
from contextlib import contextmanager
from ..errors import ConnectionError

DEFAULT_CONF_FILE = join(sep, 'etc', 'multipath.conf')

from ..connection import MessageLength, UnixDomainSocket

class MultipathClient(object):
    """ Client to the multipath-tools daemon
    """

    def __init__(self, connection=None):
        super(MultipathClient, self).__init__() #pragma: no cover
        self._connection = connection or UnixDomainSocket()

    @contextmanager
    def _with_connection_open(self):
        try:
            self._connection.connect()
            yield
        finally:
            self._connection.disconnect()

    def is_running(self):
        try:
            with self._with_connection_open():
                pass
            return True
        except ConnectionError:
            pass
        return False

    def _get_message_size_as_string(self, message):
        instance = MessageLength()
        instance.length = len(message)
        return MessageLength.write_to_string(instance)

    def _get_expected_message_size_from_string(self, string):
        return MessageLength.create_from_string(string).length

    def _send_and_receive(self, message):
        from ..model import strip_ansi_colors
        message = "%s\n" % message
        with self._with_connection_open():
            self._connection.send(self._get_message_size_as_string(message))
            self._connection.send(message)
            stream = self._connection.receive(MessageLength.min_max_sizeof().max)
            expected_length = self._get_expected_message_size_from_string(stream)
            response = self._connection.receive(expected_length)
            return strip_ansi_colors(response.strip('\x00\n'))

    def rescan(self):
        result = self._send_and_receive("reconfigure")
        if result != 'ok':
            raise RuntimeError(result) #pragma: no-cover

    def get_multipathd_conf(self):
        """ retreive multipathd.conf
        """
        from ..config import Configuration
        return Configuration.from_multipathd_conf(self._send_and_receive("show config"))

    def write_to_multipathd_conf(self, value, filepath=DEFAULT_CONF_FILE):
        """ save configuration to multipathd.conf file
        in order for the configuration to take affect, call reload()
        """
        from ..config import Configuration
        assert isinstance(value, Configuration)
        with open(filepath, 'w') as fd:
            fd.write(value.to_multipathd_conf())

    def get_list_of_multipath_devices(self):
        """ returns a list of mulipath devices and their attributes, as seen from multipathd
        """
        from ..model import get_list_of_multipath_devices_from_multipathd_output
        maps_topology = self._send_and_receive("show multipaths topology")
        paths_table = self._send_and_receive("show paths")
        result = get_list_of_multipath_devices_from_multipathd_output(maps_topology, paths_table)
        return result

    def fail_path(self, path_id):
        result = self._send_and_receive("fail path %s" % path_id)
        if result != 'ok':
            raise RuntimeError(result) #pragma: no-cover

    def reinstate_path(self, path_id):
        result = self._send_and_receive("reinstate path %s" % path_id)
        if result != 'ok': #pragma: no-cover
            raise RuntimeError(result) #pragma: no-cover

    def get_version(self):
        from ..model import parse_multipath_tools_version
        result = self._send_and_receive("?")
        version_string = parse_multipath_tools_version(result)
        return [int(item) for item in version_string.split(".")]

__all__ = ('MultipathClient')




def print_examples():
    from .tests import MOCK_OUTPUT
    from . import get_list_of_multipath_devices_from_multipathd_output
    from json import dumps
    for output in MOCK_OUTPUT:
        maps_topology = output['show multipaths topology']
        paths_table = output['show paths']
        devices = get_list_of_multipath_devices_from_multipathd_output(maps_topology, paths_table)
        print dumps(devices, sort_keys=True, indent=4)
if __name__ == "__main__":
    print_examples()




from munch import Munch

class MultipathDevice(Munch):
    def __init__(self, id, device_name, dm_name):
        super(MultipathDevice, self).__init__()
        self.path_groups = []
        self.id = id
        self.device_name = device_name
        self.dm_name = dm_name
        with open("/sys/block/{}/dev".format(dm_name)) as fd:
            self.major_minor = tuple([int(number) for number in fd.read().split(':')])

class PathGroup(Munch):
    def __init__(self, state, priority):
        super(PathGroup, self).__init__()
        self.paths = []
        self.state = state # enabled, disabled, active, undef
        self.priority = int(priority)
        # TODO add support for service-time and queue-length that were recently added
        self.load_balancing_policy = 'round-robin'

class Path(Munch):
    def __init__(self, id, device_name, major_minor, state, priority, hctl):
        super(Path, self).__init__()
        self.id = id
        self.device_name = device_name
        self.major_minor = tuple([int(item) for item in major_minor.split(":")])
        self.state = state # active, failed, or undef
        self.priority = int(priority)
        self.hctl = tuple([int(item) for item in hctl.split(":")])

PATHGROUP_STATES = ['enabled', 'disabled', 'active', 'undef']
PATH_STATES = ['active', 'failed', 'undef']



__version__ = "0.1.27.post2.ge503fb3"
__git_commiter_name__ = "Guy Rozendorn"
__git_commiter_email__ = "guy@rzn.co.il"
__git_branch__ = 'develop'
__git_remote_tracking_branch__ = 'origin/develop'
__git_remote_url__ = 'git@infinigit.infinidat.com:host/multipathtools.git'
__git_head_hash__ = 'e503fb3070b70d7f0558ad47b60c42dd642eb95f'
__git_head_subject__ = 'HIP-1237 using gevent.socket if exists'
__git_head_message__ = ''
__git_dirty_diff__ = ''


__import__("pkg_resources").declare_namespace(__name__)

""" Python bindings to Linux multipath-tools

The multipath-tools package for Linux includes a user-space monitoring tool
for the mulipath plug-in of device-mapper.
The monitor tool can be executed by two processes:
* multipath, which is executed on demand
* multipath-tools, which runs as a daemon

The two processes are independent.
Unlike what you'd expect, the 'multipath' program does not get/set properties
from the daemon's run-time configuration, it works directly with sysfs and device-mapper.

So, running 'multipath -ll' doesn't really give you the view that the daemon has,
nor does what the kernel plug-in has; it does shows you the information from sysfs,
it doesn't mean that the daemon is aware of the latest image.
Nor does 'multipath' update the daemon's configuration. That is also why this command is
some-what resources-intensive and time-consuming.

That is why we choose to do our bindings with the daemon only.
There are several advantages for this:
* We do not issue out-of-scope commands
* We always keep the daemon up-to-date
* The daemon unsures the validity of the configuration, so we don't need to worry about that

However, the communcation with the daemon is not documented and well-defined.
If you'll real the soruce-code of multipathd, you'll see it supports a ascii-based message
passing betweeen he daemon and a cleitn, which is 'multipathd -k', 
and the communication is over a unix domain socket.
This is not ideal, but its not worse than parsing the output of multipath -l.


"""

from .client import MultipathClient
from .dtypes import PATH_STATES, PATHGROUP_STATES



from infi.pyutils.lazy import cached_function
from logging import getLogger
logger = getLogger(__name__)

DEV_NONE = 0
DEV_DEVT = 1
DEV_DEVNODE = 2
DEV_DEVMAP = 3

from munch import Munch

def munch_to_multipath_conf(munch):
    self = munch
    strings = []
    for key, value in self.items():
        if value is None:
            continue
        logger.debug("%s, %s", key, value)
        strings.append(' '.join([key, value]))
    return '\n'.join(strings)

def populate_munch_from_multipath_conf_string(instance, string):
    from re import compile, DOTALL, MULTILINE #pylint: disable-msg=W0622
    re_flags = DOTALL | MULTILINE
    key_value_pattern = compile(KEY_VALUE_PATTERN, re_flags)

    for match in key_value_pattern.finditer(string):
        key, value = match.groupdict()['key'], match.groupdict()['value']
        setattr(instance, key, value)

# taken from multipath-tools/libmultipath/dict.c

def is_parameter_supported_by_libmultipath(keyword):
    from glob import glob
    from os import path
    return any([read_shared_library(shared_library).find(keyword)
                for shared_library in glob(path.sep + path.join("lib*", "*libmultipath.so*"))])

@cached_function
def read_shared_library(shared_library):
    with open(shared_library, "rb") as fd:
        return fd.read()

@cached_function
def get_supported_keywords_for_hardware_entry():
    return [item for item in ['vendor', 'product', 'revision', 'product_blacklist',
                              'path_grouping_policy', 'getuid_callout', 'path_checker',
                              'alias_prefix', 'features', 'hardware_handler', 'prio',
                              'prio_args,' 'failback', 'rr_weight', 'no_path_retry',
                              'rr_min_io', 'rr_min_io_rq', 'pg_gtimeout', 'flush_on_last_del',
                              'fast_io_fail_tmo', 'dev_loss_tmo']
            if is_parameter_supported_by_libmultipath(item)]

class HardwareEntry(Munch):
    def __init__(self, *args, **kwargs): #pylint: disable-msg=W0613
	for item in get_supported_keywords_for_hardware_entry():
           setattr(self, item, None)

@cached_function
def get_supported_keywords_for_multipath_entry():
    return [item for item in ['wwid', 'alias', 'path_grouping_policy', 'path_selector',
                              'failback', 'rr_weight', 'no_path_retry', 'rr_min_io',
                              'rr_min_io_rq', 'pg_timeout', 'flush_on_last_del',
                              'mode', 'uid', 'gid']
            if is_parameter_supported_by_libmultipath(item)]

class MultipathEntry(Munch):
    def __init__(self, *args, **kwargs): #pylint: disable-msg=W0613
        for item in get_supported_keywords_for_multipath_entry():
           setattr(self, item, None)

@cached_function
def get_supported_keywords_for_configuration():
    return [item for item in ['verbosity', 'polling_interval', 'udev_dir', 'multipath_dir',
                              'path_selector', 'path_grouping_policy', 'getuid_callout',
                              'prio', 'prio_args', 'features', 'path_checker', 'alias_prefix',
                              'failback' , 'rr_min_io', 'rr_min_io_rq', 'max_fds', 'no_path_retry',
                              'queue_without_daemon', 'checker_timeout', 'pg_timeout',
                              'flush_on_last_del', 'user_friendly_names', 'mode', 'uid', 'gid',
                              'fast_io_fail_tmo' , 'dev_loss_tmo']
            if is_parameter_supported_by_libmultipath(item)]

class ConfigurationAttributes(Munch):
    def __init__(self, *args, **kwargs): #pylint: disable-msg=W0613
	for item in get_supported_keywords_for_configuration():
           setattr(self, item, None)

class RuleList(object):
    """ container for {black,white}list configuration:

    devnode    a list of the the regular experssions
    device     a list of the device declarations
    wwid       a list of the WWIDs being included

    """
    def __init__(self):
        super(RuleList, self).__init__()
        self.device = []
        self.devnode = []
        self.wwid = []

    def to_multipathd_conf(self):
        strings = []
        for devnode in self.devnode:
            strings.append('devnode %s' % devnode)
        for device in self.device:
            strings.append('device {')
            strings.extend(['\t%s' % line for line in munch_to_multipath_conf(device).splitlines()])
            strings.append('}')
        for wwid in self.wwid:
            strings.append('wwid %s' % wwid)
        return '\n'.join(strings)

    @classmethod
    def from_multipathd_conf(cls, string):
        from re import compile, DOTALL, MULTILINE #pylint: disable-msg=W0622
        re_flags = DOTALL | MULTILINE
        rulelist_pattern = compile(RULELIST_PATTERN, re_flags)

        instance = cls()

        for match in rulelist_pattern.finditer(string):
            key, value = match.groupdict()['key'], match.groupdict()['value']
            if key == 'devnode':
                instance.devnode.append(value)
            elif key == 'wwid':
                instance.wwid.append(value)
            elif key == 'device':
                device = Device()
                populate_munch_from_multipath_conf_string(device, value.strip("{}"))
                instance.device.append(device)
            else:
                raise AssertionError("key %s is invalid", key) # pragma: no cover
        return instance

class Device(Munch):
    def __init__(self, *args, **kwargs): #pylint: disable-msg=W0613
        for item in ['vendor', 'product']:
            setattr(self, item, None)

KEY_VALUE_PATTERN = \
    r"^\t*(?P<key>[A-Za-z0-9_]+) (?P<value>[^{}\n]+)$"

DEVICE_PATTERN = \
    r"^\t*(?P<key>device|multipath) {(?P<value>[^{}]*)}$"

RULELIST_PATTERN = \
    r"^\t*(?P<key>[A-Za-z0-9_]+) (?P<value>(?:[^{}\n]+)|({[^{}].*?}))$"

MULTIPATH_CONF_PATTERN = \
    r"^(?P<name>defaults|blacklist_exceptions|blacklist|devices|multipaths)" + \
    " {\n*(?P<content>(:?%s)*|(:?%s)*|(:?%s)*|\b|\B)\n*}\n*" % \
        (r"^[^{}\n]*$\n", # key-value pairs
         r"^\t*(?:device|multipath) {[^{}]*}$\n", # devices
         r"(?:^[^{}]*$\n)|(?:^\t*(?:device|multipath) {[^{}]*}$\n)", # both
         )

"""
        (r"^[^{}]*$\n", # key-value pairs
         r"^\t*(?:device|multipath) {[^{}]*}$\n", # devices
         r"(?:^[^{}]*$\n)|(?:^\t*(?:device|multipath) {[^{}]*}$\n)", # both
         )
"""
class Configuration(object):
    """ container for multipathd.conf

    the pupose of the class (and module) is to hold the configuration of multipathd.conf
    it provides a munch-like interface to all the configuration, where the keys are the option names as defined
    in multipath.conf.annotated.
    all values are stored as strings.

    The configuration contains several elements:

    devices        a list of devices, the devices {} section of multipath.conf
    multipaths     a list of mulltipath(es), the multipaths section of multipath.conf
    blacklist      the blacklist section of multipath.conf
    whitelist      the blacklist_exceptions section of multipath.conf

    This class also provides two-way translation to/from the syntax of multipath.conf.
    See the help of to_multipathd_conf and from_multipathd_conf
    """

    def __init__(self):
        super(Configuration, self).__init__()
        self.attributes = ConfigurationAttributes()
        self.devices = []
        self.multipaths = []
        self.blacklist = RuleList()
        self.whitelist = RuleList()

    def to_multipathd_conf(self):
        """ return a string representation of configuration is the syntax of multipath.conf
        unused values are not written to the configuration file
        """
        strings = []
        strings.append('defaults {')
        strings.extend(['\t%s' % line for line in munch_to_multipath_conf(self.attributes).splitlines()])
        strings.append('}')
        strings.append('blacklist {')
        strings.extend(['\t%s' % line for line in self.blacklist.to_multipathd_conf().splitlines()])
        strings.append('}')
        strings.append('blacklist_exceptions {')
        strings.extend(['\t%s' % line for line in self.whitelist.to_multipathd_conf().splitlines()])
        strings.append('}')
        strings.append('devices {')
        for device in self.devices:
            strings.append('\tdevice {')
            strings.extend(['\t\t%s' % line for line in munch_to_multipath_conf(device).splitlines()])
            strings.append('\t}')
        strings.append('}')
        strings.append('multipaths {')
        for multipath in self.multipaths:
            strings.append('\tmultipath {')
            strings.extend(['\t\t%s' % line for line in munch_to_multipath_conf(multipath).splitlines()])
            strings.append('\t}')
        strings.append('}')
        return '\n'.join(strings)

    @classmethod
    def from_multipathd_conf(cls, show_config_output):
        """ this class method takes the multipathd configuration, as taken from multipathd -k; show config
        and returns an instance with all the configuration parsed out of the input

        this method may fail parsing the configuration if it was read directly from the configuration file,
        since it makes some assumpsions that are always valid when looking at the output of 'show config' command.
        For exmaple:
        * identations are with tab characters, and not spaces
        * between and key and value, only one space character exists
        """
        from re import compile, DOTALL, MULTILINE #pylint: disable-msg=W0622
        re_flags = DOTALL | MULTILINE
        pattern = compile(MULTIPATH_CONF_PATTERN, re_flags)
        device_pattern = compile(DEVICE_PATTERN, re_flags)
        instance = cls()

        def populate_list(list, base_class, content): #pylint: disable-msg=W0622
            for match in device_pattern.finditer(content):
                _, value = match.groupdict()['key'], match.groupdict()['value']
                entry = base_class()
                populate_munch_from_multipath_conf_string(entry, value)
                list.append(entry)

        for section in pattern.finditer(show_config_output):
            name, content = section.groupdict()['name'], section.groupdict()['content']
            if name == 'defaults':
                populate_munch_from_multipath_conf_string(instance.attributes, content)
            elif name == 'blacklist':
                instance.blacklist = RuleList.from_multipathd_conf(content)
            elif name == 'blacklist_exceptions':
                instance.whitelist = RuleList.from_multipathd_conf(content)
            elif name == 'devices':
                populate_list(instance.devices, HardwareEntry, content)
            elif name == 'multipaths':
                populate_list(instance.multipaths, MultipathEntry, content)

        return instance






class ServiceException(Exception):
    pass

class ServiceFailedToStop(ServiceException):
    pass

class ServiceFailedToStart(ServiceException):
    pass

class EntryPoint(object):
    """ This base class is for stuff that needs to exists for multipath to work
    """
    def __init__(self):
        object.__init__(self)

    def is_installed(self):
        raise NotImplementedError # pragma: no cover

    def is_running(self):
        raise NotImplementedError # pragma: no cover

    def stop(self):
        raise NotImplementedError # pragma: no cover

    def start(self):
        raise NotImplementedError # pragma: no cover

    def reload(self):
        raise NotImplementedError # pragma: no cover

    def restart(self):
        raise NotImplementedError # pragma: no cover

def get_list_of_live_modules():
    from os.path import join, sep
    with open(join(sep, 'proc', 'modules')) as modules:
        return [line.split()[0] for line in modules.readlines()]

def execute(command):
    from logging import getLogger
    log = getLogger(__name__)
    debug = log.debug
    from infi.execute import execute as _execute
    debug('executing %s', command)
    try:
        subprocess = _execute(command.split())
    except:
        from infi.exceptools import chain
        raise chain(ServiceException)
    debug('waiting for it')
    subprocess.wait()
    debug('returncode = %s', subprocess.get_returncode())
    debug('stdout = %s', subprocess.get_stdout())
    debug('stderr = %s', subprocess.get_stderr())
    if subprocess.get_returncode() != 0:
        raise ServiceException(subprocess.get_stderr())

def get_kernel_release():
    import platform
    (_, _, release, _, _, _) = platform.uname()
    return release

def get_modules_list():
    from os.path import sep, join
    modules = []
    with open(join(sep, 'lib', 'modules', get_kernel_release(), 'modules.dep')) as modules_dep:
        content = modules_dep.read()
        for line in content.splitlines():
            module_path, _ = line.split(':', 1)
            module_name = module_path.split(sep)[-1].split(".")[0]
            modules.append(module_name)
    return modules

class KernelModule(EntryPoint):
    def __init__(self, module_name):
        super(KernelModule, self).__init__()
        self.module_name = module_name

    def is_installed(self):
        return self.module_name in get_modules_list()

    def is_running(self):
        return self.module_name in get_list_of_live_modules()

    def stop(self):
        try:
            execute('rmmod %s' % self.module_name)
        except ServiceException:
            from infi.exceptools import chain
            raise chain(ServiceFailedToStop)

    def start(self):
        try:
            execute('modprobe %s' % self.module_name)
        except ServiceException:
            from infi.exceptools import chain
            raise chain(ServiceFailedToStart)

    def reload(self):
        pass # pragma: no cover

    def restart(self):
        pass # pragma: no cover

class File(EntryPoint):
    def __init__(self, path):
        super(File, self).__init__()
        self.path = path

    def is_installed(self):
        from os.path import exists
        return exists(self.path)

    def is_running(self):
        return True

    def stop(self):
        pass

    def start(self):
        pass

    def reload(self):
        pass

    def restart(self):
        pass

def get_list_of_running_processes():
    from os import listdir, readlink
    from os.path import exists, join, sep

    processes = []
    for item in listdir(join(sep, 'proc')):
        if not item.isdigit():
            continue
        if not exists(join(sep, 'proc', item, 'exe')):
            continue
        executable = readlink(join(sep, 'proc', item, 'exe')).split(sep)[-1]
        processes.append(executable)
    return processes

class InitScript(EntryPoint):
    def __init__(self, script_name, process_name=None):
        super(InitScript, self).__init__()
        self.script_name = script_name
        self.process_name = process_name if process_name is not None else script_name

    def is_installed(self):
        from os.path import exists
        return exists(self._get_script_path())

    def is_running(self):
        from os import listdir
        from os.path import exists, join, sep
        from logging import getLogger
        logger = getLogger(__name__)
        debug = logger.debug
        pid_file = join(sep, 'var', 'run', '%s.pid' % self.process_name)
        if not exists(pid_file):
            debug("did not found pid file %s", pid_file)
            return False
        with open(pid_file, 'r') as fd:
            pid = fd.read().splitlines()[0].strip()
            pid_running = pid in listdir(join(sep, 'proc'))
            debug("is pid %s listed under /proc? %s", pid, pid_running)
            return pid_running

    def _get_script_path(self):
        from os.path import sep, join
        return join(sep, 'etc', 'init.d', self.script_name)

    def _execute(self, command):
        execute("%s %s" % (self._get_script_path(), command))

    def stop(self):
        self._execute('stop')

    def start(self):
        self._execute('start')

    def reload(self):
        self._execute('reload')

    def restart(self):
        self._execute('restart')

class CompositeEntryPoint(EntryPoint):
    def __init__(self):
        EntryPoint.__init__(self)
        self.components = {}

    def add_component(self, component, load_order):
        assert(isinstance(component, EntryPoint))
        assert(isinstance(load_order, int))

        if component in self.components.values():
            return
        self.components[load_order] = component

    def remove_component(self, component):
        assert(isinstance(component, EntryPoint))

        if component not in self.components.values():
            return

        for key, value in self.components.items():
            if value is component:
                self.components.pop(key)

    def iter_components_by_order(self, increasing_order=True):
        keys = self.components.keys()
        keys.sort()
        if not increasing_order:
            keys.reverse()
        for key in keys:
            yield key, self.components[key]

    def is_installed(self):
        for component in self.components.values():
            if not component.is_installed():
                return False
        return True

    def is_running(self):
        for component in self.components.values():
            if not component.is_running():
                return False
        return True

    def stop(self):
        for _, value in self.iter_components_by_order(False):
            if value.is_running():
                value.stop()

    def start(self):
        for _, value in self.iter_components_by_order(True):
            if not value.is_running():
                value.start()

    def restart(self):
        self.stop()
        self.start()

    def reload(self):
        for _, value in self.iter_components_by_order(True):
            value.reload()

def get_multipath_composite():
    composite = CompositeEntryPoint()
    for item in ['multipath', 'dm-multipath']:
        component = KernelModule(item)
        if component.is_installed():
            break
    composite.add_component(component, 10)
    for item in ['multipathd', 'multipath-tools']:
        component = InitScript(item, 'multipathd')
        if component.is_installed():
            break
    composite.add_component(component, 20)
    return composite




from unittest import TestCase, SkipTest
from mock import patch, Mock, MagicMock

#pylint: disable-all

EMPTY_CONFIGURATION = """
defaults {
}
blacklist {
}
blacklist_exceptions {
}
devices {
}
multipaths {
}
""".strip('\n')

EMPTY_CONFIGURATION__EMPTY_CHILDREN = """
defaults {
}
blacklist {
    device {
    }
}
blacklist_exceptions {
    device {
    }
}
devices {
    device {
    }
}
multipaths {
    multipath {
    }
}
""".strip('\n').replace('    ', '\t')

KEY_VALUE_ITEMS = """
a 1
b "hello 1"
foo legen...|wait.for\wit|[^dary]
""".strip('\n').replace('    ', '\t')

SAMPLE_CONFIGURATIONS = [
    'defaults {\n\tverbosity 2\n\tpolling_interval 5\n\tudev_dir "/dev"\n\tmultipath_dir "/lib64/multipath"\n\tpath_selector "round-robin 0"\n\tpath_grouping_policy failover\n\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\tprio const\n\tfeatures "0"\n\tpath_checker directio\n\tfailback manual\n\trr_min_io 1000\n\trr_weight uniform\n\tqueue_without_daemon yes\n\tpg_timeout none\n\tflush_on_last_del no\n\tuser_friendly_names yes\n\tfind_multipaths no\n\tlog_checker_err always\n}\nblacklist {\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevnode "^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*"\n\tdevnode "^hd[a-z]"\n\tdevnode "^dcssblk[0-9]*"\n\tdevice {\n\t\tvendor "STK"\n\t\tproduct "Universal Xport"\n\t}\n\tdevice {\n\t\tvendor "DGC"\n\t\tproduct "LUNZ"\n\t}\n\tdevice {\n\t\tvendor "EMC"\n\t\tproduct "LUNZ"\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "S/390.*"\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "S/390.*"\n\t}\n\tdevice {\n\t\tvendor "STK"\n\t\tproduct "Universal Xport"\n\t}\n}\nblacklist_exceptions {\n}\ndevices {\n\tdevice {\n\t\tvendor "NEXSAN"\n\t\tproduct "SATABoy2"\n\t}\n\tdevice {\n\t\tvendor "COMPELNT"\n\t\tproduct "Compellent Vol"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "APPLE*"\n\t\tproduct "Xserve RAID "\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "3PARdata"\n\t\tproduct "VV"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "DEC"\n\t\tproduct "HSG80"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker hp_sw\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "1 hp_sw"\n\t\tprio hp_sw\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "HP"\n\t\tproduct "A6189A"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\tno_path_retry 12\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "(COMPAQ|HP)"\n\t\tproduct "(MSA|HSV)1.0.*"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker hp_sw\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "1 hp_sw"\n\t\tprio hp_sw\n\t\trr_weight uniform\n\t\tno_path_retry 12\n\t\trr_min_io 100\n\t}\n\tdevice {\n\t\tvendor "(COMPAQ|HP)"\n\t\tproduct "MSA VOLUME"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 12\n\t\trr_min_io 100\n\t}\n\tdevice {\n\t\tvendor "(COMPAQ|HP)"\n\t\tproduct "HSV1[01]1|HSV2[01]0|HSV300|HSV4[05]0"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 12\n\t\trr_min_io 100\n\t}\n\tdevice {\n\t\tvendor "HP"\n\t\tproduct "MSA2[02]12fc|MSA2012i"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 18\n\t\trr_min_io 100\n\t}\n\tdevice {\n\t\tvendor "HP"\n\t\tproduct "MSA2012sa|MSA23(12|24)(fc|i|sa)|MSA2000s VOLUME"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 18\n\t\trr_min_io 100\n\t}\n\tdevice {\n\t\tvendor "HP"\n\t\tproduct "HSVX700"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "1 alua"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 12\n\t\trr_min_io 100\n\t}\n\tdevice {\n\t\tvendor "HP"\n\t\tproduct "LOGICAL VOLUME.*"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\tno_path_retry 12\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "HP"\n\t\tproduct "OPEN-.*"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\tno_path_retry 18\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "HP"\n\t\tproduct "P2000 G3 FC|P2000G3 FC/iSCSI|P2000 G3 SAS|P2000 G3 iSCSI"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 18\n\t\trr_min_io 100\n\t}\n\tdevice {\n\t\tvendor "DDN"\n\t\tproduct "SAN DataDirector"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "EMC"\n\t\tproduct "SYMMETRIX"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --page=pre-spc3-83 --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\tno_path_retry 6\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "DGC"\n\t\tproduct ".*"\n\t\tproduct_blacklist "LUNZ"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker emc_clariion\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "1 emc"\n\t\tprio emc\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 60\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "EMC"\n\t\tproduct "Invista"\n\t\tproduct_blacklist "LUNZ"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\tno_path_retry 5\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "FSC"\n\t\tproduct "CentricStor"\n\t\tpath_grouping_policy group_by_serial\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "HITACHI"\n\t\tproduct "OPEN-.*"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\tno_path_retry 6\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "HITACHI"\n\t\tproduct "DF.*"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio hds\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "ProFibre 4000R"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1722-600"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 300\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1724"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 300\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1726"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 300\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1742"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1745|1746"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "2 pg_init_retries 50"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 15\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1745"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "2 pg_init_retries 50"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 15\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1746"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "2 pg_init_retries 50"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 15\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1814"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1815"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1818"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "3526"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "3542"\n\t\tpath_grouping_policy group_by_serial\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "2105800"\n\t\tpath_grouping_policy group_by_serial\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "2105F20"\n\t\tpath_grouping_policy group_by_serial\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "1750500"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "2107900"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "2145"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "S/390 DASD ECKD"\n\t\tproduct_blacklist "S/390.*"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/sbin/dasdinfo -u -b %n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "S/390 DASD FBA"\n\t\tproduct_blacklist "S/390.*"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/sbin/dasdinfo -u -b %n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "IPR.*"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "1 alua"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "AIX"\n\t\tproduct "VDASD"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 60\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "3303      NVDISK"\n\t\tpath_grouping_policy failover\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 60\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "AIX"\n\t\tproduct "NVDISK"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "1 alua"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 60\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "DELL"\n\t\tproduct "MD3000"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "2 pg_init_retries 50"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 15\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "DELL"\n\t\tproduct "MD3000i"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "2 pg_init_retries 50"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 15\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "DELL"\n\t\tproduct "MD32xx"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "2 pg_init_retries 50"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 15\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "DELL"\n\t\tproduct "MD32xxi"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "2 pg_init_retries 50"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 15\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "NETAPP"\n\t\tproduct "LUN.*"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio ontap\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\trr_min_io 128\n\t}\n\tdevice {\n\t\tvendor "IBM"\n\t\tproduct "Nseries.*"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio ontap\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\trr_min_io 128\n\t}\n\tdevice {\n\t\tvendor "Pillar"\n\t\tproduct "Axiom.*"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio alua\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "SGI"\n\t\tproduct "TP9[13]00"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "SGI"\n\t\tproduct "TP9[45]00"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "SGI"\n\t\tproduct "IS.*"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "2 pg_init_retries 50"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 15\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "STK"\n\t\tproduct "OPENstorage D280"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "SUN"\n\t\tproduct "(StorEdge 3510|T4)"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "PIVOT3"\n\t\tproduct "RAIGE VOLUME"\n\t\tpath_grouping_policy multibus\n\t\tgetuid_callout "/lib/udev/scsi_id --page=0x80 --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "1 queue_if_no_path"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 100\n\t}\n\tdevice {\n\t\tvendor "SUN"\n\t\tproduct "CSM200_R"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "SUN"\n\t\tproduct "LCSM100_[IEFS]"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "(LSI|ENGENIO)"\n\t\tproduct "INF-01-00"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "2 pg_init_retries 50"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry 15\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "STK"\n\t\tproduct "FLEXLINE 380"\n\t\tproduct_blacklist "Universal Xport"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker rdac\n\t\tfeatures "0"\n\t\thardware_handler "1 rdac"\n\t\tprio rdac\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\tno_path_retry queue\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "EUROLOGC"\n\t\tproduct "FC2502"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --page=0x80 --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker directio\n\t\tfeatures "0"\n\t\thardware_handler "0"\n\t\tprio const\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n\tdevice {\n\t\tvendor "NEC"\n\t\tproduct "DISK ARRAY"\n\t\tpath_grouping_policy group_by_prio\n\t\tgetuid_callout "/lib/udev/scsi_id --whitelisted --device=/dev/%n"\n\t\tpath_selector "round-robin 0"\n\t\tpath_checker tur\n\t\tfeatures "0"\n\t\thardware_handler "1 alua"\n\t\tprio alua\n\t\tfailback immediate\n\t\trr_weight uniform\n\t\trr_min_io 1000\n\t}\n}\nmultipaths {\n}',
                         ]

class OutputTests(TestCase):
    def test_output__empty_configuration(self):
        from . import Configuration
        config = Configuration()
        self.assertEquals(config.to_multipathd_conf(), EMPTY_CONFIGURATION)

    def test_output__empty_configuration__everything_is_none(self):
        from . import Configuration, Device, MultipathEntry, HardwareEntry
        config = Configuration()
        config.blacklist.device.append(Device())
        config.whitelist.device.append(Device())
        config.multipaths.append(HardwareEntry())
        config.devices.append(MultipathEntry())
        self.assertEquals(config.to_multipathd_conf(), EMPTY_CONFIGURATION__EMPTY_CHILDREN)

from re import compile, DOTALL, MULTILINE
from . import KEY_VALUE_PATTERN, MULTIPATH_CONF_PATTERN

from os.path import exists, join, sep, abspath, dirname
from pkg_resources import resource_filename
SAMPLE_FILEPATH = resource_filename(__name__, "sample.txt")


class InputTests(TestCase):
    def test_empty_configuration(self):
        from . import Configuration
        config = Configuration.from_multipathd_conf(EMPTY_CONFIGURATION)

    def test_samples(self):
        from . import Configuration
        pattern = compile(MULTIPATH_CONF_PATTERN, DOTALL | MULTILINE)
        for sample in SAMPLE_CONFIGURATIONS:
            matches = pattern.match(sample)
            config = Configuration.from_multipathd_conf(sample)

    def test_regex__empty_configuration(self):
        pattern = compile(MULTIPATH_CONF_PATTERN, DOTALL | MULTILINE)
        matches = [match.groupdict() for match in pattern.finditer(EMPTY_CONFIGURATION)]
        sections = [dict['name'] for dict in matches]
        self.assertIn('defaults', sections)
        self.assertIn('blacklist', sections)
        self.assertIn('blacklist_exceptions', sections)
        self.assertIn('devices', sections)
        self.assertIn('multipaths', sections)

    def test_regex__key_value_pattern(self):
        pattern = compile(KEY_VALUE_PATTERN, DOTALL | MULTILINE)
        result = pattern.search("key value")
        self.assertEquals(result.groupdict()['key'], 'key')
        self.assertEquals(result.groupdict()['value'], 'value')
        result = pattern.search(r"regex ^this|is|[a][^complicated](regex).*/")
        self.assertEquals(result.groupdict()['key'], 'regex')
        self.assertEquals(result.groupdict()['value'], '^this|is|[a][^complicated](regex).*/')

    def test_regex__multiple_key_values(self):
        pattern = compile(KEY_VALUE_PATTERN, DOTALL | MULTILINE)
        iterator = pattern.finditer(KEY_VALUE_ITEMS)

        expected_keys = ['a', 'b', 'foo']
        expected_values = ['1', '"hello 1"', "legen...|wait.for\wit|[^dary]"]
        for match in iterator:
            self.assertEquals(len(match.groupdict().keys()), 2)
            self.assertIn(match.groupdict()['key'], expected_keys)
            self.assertIn(match.groupdict()['value'], expected_values)

    def test_populate_munch(self):
        from . import populate_munch_from_multipath_conf_string
        from munch import Munch
        string = """
            foo bar
            high lower
            1 0
            """.replace('    ', '\t')
        instance = Munch()
        populate_munch_from_multipath_conf_string(instance, string)
        self.assertEquals(instance.foo, 'bar')
        self.assertEquals(instance.high, 'lower')
        self.assertEquals(instance['1'], '0')

    def test_rulelist_from_string(self):
        from . import RuleList
        string = """
        devnode ^something%N
        devnode else$
        wwid 010203040506
        device {
            product someProduct
            vendor someVendor
        }
        """.replace('    ', '\t')
        instance = RuleList.from_multipathd_conf(string)
        self.assertEqual(len(instance.device), 1)
        self.assertEqual(len(instance.devnode), 2)
        self.assertEqual(len(instance.wwid), 1)
        self.assertEquals(instance.device[0].product, 'someProduct')
        self.assertEquals(instance.device[0].vendor , 'someVendor')
        self.assertEquals(instance.devnode, ['^something%N', 'else$'])
        self.assertEquals(instance.wwid, ['010203040506', ])


    def test_regex__sample_configuration(self):
        assert exists(SAMPLE_FILEPATH)
        with open(SAMPLE_FILEPATH, 'r') as fd:
            sample_config_string = fd.read()
        assert len(sample_config_string)

        pattern = compile(MULTIPATH_CONF_PATTERN, DOTALL | MULTILINE)
        matches = [match.groupdict() for match in pattern.finditer(sample_config_string)]
        sections = [dict['name'] for dict in matches]
        self.assertIn('defaults', sections)
        self.assertIn('blacklist', sections)
        self.assertIn('blacklist_exceptions', sections)
        self.assertIn('devices', sections)
        self.assertIn('multipaths', sections)

    def test_sample_configuration(self):
        from . import Configuration

        assert exists(SAMPLE_FILEPATH)
        with open(SAMPLE_FILEPATH, 'r') as fd:
            sample_config_string = fd.read()
        assert len(sample_config_string)

        config = Configuration.from_multipathd_conf(sample_config_string)
        self.assertEqual(config.attributes.udev_dir, '/dev')
        self.assertEqual(config.attributes.polling_interval, '10')
        self.assertEquals(config.attributes.uid, '0')
        self.assertEqual(config.blacklist.wwid, ['26353900f02796769'])
        self.assertIn('"^dcssblk[0-9]*"', config.blacklist.devnode)
        self.assertEqual(config.blacklist.device[0].product, "MSA[15]00")
        self.assertIn('"IBM.75000000092461.4d00.36"', config.whitelist.wwid)
        self.assertEqual(len(config.devices), 2)
        self.assertEqual(config.multipaths[0].wwid, "3600508b4000156d700012000000b0000")
        self.assertEqual(config.multipaths[0].failback, "manual")
        self.assertEqual(len(config.multipaths), 2)
        self.assertEqual(config.devices[1].path_grouping_policy, "multibus")

class InOut(TestCase):
    def test_sample_configuration(self):
        raise SkipTest
        from . import Configuration

        assert exists(SAMPLE_FILEPATH)
        with open(SAMPLE_FILEPATH, 'r') as fd:
            sample_config_string = fd.read()
        assert len(sample_config_string)
        config = Configuration.from_multipathd_conf(sample_config_string)
        self.assertEqual(config.to_multipathd_conf(), sample_config_string)




from munch import Munch
from re import compile, MULTILINE, DOTALL

from logging import getLogger
logger = getLogger(__name__)

HCTL = r'\d+:\d+:\d+:\d+'
DEV = r'\w+'
DEV_T = r'\d+:\d+'
PRI = r'\d+'
DM_ST = r'(?:\[\w+\])|(?:\w+)'
CHK_ST = DM_ST
NEXT_CHECK = '(?:[X\. /0-9]+)|orphan *|\[orphan\] *'


def parse_paths_table(paths_table):
    PATH_PATTERN = r"^" + \
            ("(?P<hctl>%s) +(?P<dev>%s) +" % (HCTL, DEV)) + \
            (r"(?P<dev_t>%s) +(?P<pri>%s) +" % (DEV_T, PRI)) + \
            (r"(?P<dm_st>%s) *(?P<chk_st>%s) +" % (DM_ST, CHK_ST)) + \
             r"(?:\b|\B|\w+ +)" + \
            (r"(?P<next_check>%s)$" % (NEXT_CHECK,))
    pattern = compile(PATH_PATTERN, MULTILINE | DOTALL)
    matches = []
    for match in pattern.finditer(paths_table):
        logger.debug("match found: %s", match.groupdict())
        matches.append(dict((key, value.strip('[]')) for (key, value) in match.groupdict().items()))
    return matches


def parse_multipaths_topology(maps_topology):
    MULTIPATH_PATTERN = r"^" + \
        r"(?:(?P<action>\w+): +|\b|\B)" + \
        r"(?:(?P<alias>[A-Za-z0-9_]+) +|\b|\B)" + \
        r"(?P<wwid>[A-Za-z0-9_\(\) ]+) *" + \
        r"(?P<dm>dm-\w+) +" \
        r"(?P<vendor>\w+) *,(?P<product>\w+)(?P<rev>.*)\n" + \
        r"(?P<options>.+)\n" + \
        r"(?P<path_groups>(?:(?:.*[\|`\\].*\n?)*))"
    pattern = compile(MULTIPATH_PATTERN, MULTILINE)
    matches = []
    for match in pattern.finditer(maps_topology):
        logger.debug("multipath found: %s", match.groupdict())
        multipath_dict = dict((key, value.strip(' ()[]') if value is not None else value) \
                              for (key, value) in match.groupdict().items())
        parse_path_groups_in_multipath_dict(multipath_dict)
        matches.append(multipath_dict)
    return matches

def parse_path_groups_in_multipath_dict(multipath_dict):
    PATH_GROUP_PATTERN = r"^" + \
        r"[^\n]*" + \
        r"\[?prio=(?P<prio>\d+)\]?" + \
        r" *" + \
        r"(?:status=|\b|\B)" + \
        r"\[?(?P<state>\w+)\]?" + \
        r" *\n" + \
        r"(?P<paths>(?:(?:.*:.*\n?)*))"
    pattern = compile(PATH_GROUP_PATTERN, MULTILINE)
    matches = []
    for match in pattern.finditer(multipath_dict['path_groups']):
        logger.debug("pathgroup found: %s", match.groupdict())
        pathgroup_dict = dict((key, value.strip(' ()[]') if value is not None else value) for (key, value) in match.groupdict().items())
        parse_paths_in_pathgroup_dict(pathgroup_dict)
        matches.append(pathgroup_dict)
    multipath_dict['path_groups'] = matches

def parse_paths_in_pathgroup_dict(pathgroup_dict):
    PATH_PATTERN = r"(?P<hctl>%s) +(?P<dev>%s) +(?P<dev_t>%s)" % (HCTL, DEV, DEV_T)
    pattern = compile(PATH_PATTERN, MULTILINE | DOTALL)
    matches = []
    for match in pattern.finditer(pathgroup_dict['paths']):
        logger.debug("path found: %s", match.groupdict())
        matches.append(dict((key, value.strip(' []')) for (key, value) in match.groupdict().items()))
    pathgroup_dict['paths'] = matches

def dict_by_attribute(attr_name, list_obj):
    result = dict()
    for item in list_obj:
        result[item[attr_name]] = item
    return result

def get_list_of_multipath_devices_from_multipathd_output(maps_topology, paths_table):
    from ..dtypes import MultipathDevice, Path, PathGroup
    multipaths = parse_multipaths_topology(maps_topology)
    paths = parse_paths_table(paths_table)
    paths_by_mjmn = dict_by_attribute('dev_t', paths)
    result = []
    logger.debug("multipaths = %s", multipaths)
    for mpath_dict in multipaths:
        try:
            multipath = MultipathDevice(mpath_dict['wwid'], mpath_dict['alias'] or mpath_dict['wwid'], mpath_dict['dm'])
        except IOError, err:
            logger.error("MultipathDevice disappeared: {}".format(err))
            continue
        for pathgroup_dict in mpath_dict['path_groups']:
            path_group = PathGroup(pathgroup_dict['state'], pathgroup_dict['prio'])
            for path_dict in pathgroup_dict['paths']:
                mjmn = path_dict['dev_t']
                if mjmn not in paths_by_mjmn.keys():
                    logger.debug("There is no path for major:minor {}, only for {}".format(mjmn, paths_by_mjmn))
                    continue
                path_info = paths_by_mjmn[mjmn]
                path = Path(path_info['dev'], path_info['dev'], path_info['dev_t'],
                            path_info['dm_st'], path_info['pri'], path_info['hctl'])
                path_group.paths.append(path)
            multipath.path_groups.append(path_group)
        result.append(multipath)
    return result

def strip_ansi_colors(string):
    pattern = compile(r"\x1b\[[;\d]*[A-Za-z]", MULTILINE | DOTALL)
    return pattern.sub("", string)

def parse_multipath_tools_version(string):
    pattern = compile(r"^multipath-tools v(?P<version>[0-9\.]+)", MULTILINE)
    match = pattern.search(string)
    return match.groupdict()['version']




def print_config():
    from ..client import MultipathClient
    client = MultipathClient()
    print client._send_and_receive("show config")

def print_maps():
    from ..client import MultipathClient
    client = MultipathClient()
    print client._send_and_receive("show multipaths topology")
    print client._send_and_receive("show paths")




#pylint: disable-all

import unittest
import mock

from contextlib import contextmanager, nested
from time import sleep
from . import EntryPoint

class MockEntryPoint(EntryPoint):
    def __init__(self):
        EntryPoint.__init__(self)
        self._running = False
        self._installed = True

    def is_installed(self):
        return self._installed

    def is_running(self):
        return self._running

    def stop(self):
        self._running = False

    def start(self):
        self._running = True

    def restart(self):
        pass

    def reload(self):
        pass

class TestCase(unittest.TestCase):
    def setUp(self):
        if self.should_skip():
            raise unittest.SkipTest

    def tearDown(self):
        pass

    def should_skip(self):
        return True

    def _execute_side_effect(self, return_value):
        execute = mock.Mock()
        execute.wait = mock.Mock()
        execute.get_returncode = mock.Mock()
        execute.get_returncode.return_value = return_value
        execute.get_stdout = mock.Mock()
        execute.get_stdout.return_value = ''
        execute.get_stderr = mock.Mock()
        execute.get_stderr.return_value = ''

        def side_effect(*args, **kwargs):
            return execute

        return side_effect

    @contextmanager
    def mock_execute(self, return_value):
        with mock.patch("infi.execute.execute") as execute:
            execute.side_effect = self._execute_side_effect(return_value)
            yield execute

class CompositeServiceTestCase(TestCase):
    def test_get_composite(self):
        from . import get_multipath_composite
        composite = get_multipath_composite()
        self.assertTrue(composite.is_installed())

class FileTestCase(unittest.TestCase):
    def setUp(self):
        from tempfile import mkstemp
        from os import close
        fd, self.tempfile = mkstemp()
        close(fd)

    def test_file_that_does_not_exist(self):
        from . import File
        item = File('abracadabra')
        self.assertFalse(item.is_installed())

    def test_file_that_exists(self):
        from . import File
        item = File(self.tempfile)
        self.assertTrue(item.is_installed())
        self.assertTrue(item.is_running())
        item.stop()
        item.start()
        item.reload()
        item.restart()

    def tearDown(self):
        from os import remove
        remove(self.tempfile)

class InitServiceTestCase(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self._choose_service()

    def _choose_service(self):
        from os.path import exists, sep, join
        for item in ['ssh', 'sshd']:
            if exists(join(sep, 'etc', 'init.d', item)):
                self.service_name = item
                return
        raise unittest.SkipTest

    def test_get_list_of_running_processes(self):
        from . import get_list_of_running_processes
        actual = get_list_of_running_processes()
        self.assertGreater(len(actual), 0)

    def test_is_installed__service_does_not_exist(self):
        from . import InitScript
        item = InitScript("foo")
        self.assertFalse(item.is_installed())

    def test_is_installed(self):
        from . import InitScript
        item = InitScript(self.service_name, 'sshd')
        self.assertTrue(item.is_installed())

    def test_is_running(self):
        from . import InitScript
        item = InitScript(self.service_name, 'sshd')
        self.assertTrue(item.is_running())

    def test_start(self):
        from . import InitScript
        item = InitScript(self.service_name, 'sshd')
        if not item.is_running():
            item.stop()
        item.start()
        self.assertTrue(item.is_running())

    def test_stop(self):
        from . import InitScript
        item = InitScript(self.service_name, 'sshd')
        if item.is_running():
            item.start()
        item.stop()
        self.assertFalse(item.is_running())
        item.start()

    def test_restart(self):
        from . import InitScript
        item = InitScript(self.service_name, 'sshd')
        item.restart()

    def test_reload(self):
        from . import InitScript
        item = InitScript(self.service_name, 'sshd')
        item.reload()

class KernelModuleTestCase(TestCase):
    def _get_module_in_use(self):
        from os.path import join, sep, exists
        with open(join(sep, 'proc', 'modules')) as modules:
            for line in modules.readlines():
                module_name, _, used_by_counter, _, _, _ = line.split()
                if int(used_by_counter):
                    return module_name
        return None

    def test_is_installed(self):
        from . import KernelModule
        item = KernelModule('floppy')
        self.assertTrue(item.is_installed())

    def test_is_installed__no_such_module(self):
        from . import KernelModule
        item = KernelModule('thereisNoSuchModule')
        self.assertFalse(item.is_installed())

    def test_start(self):
        from . import KernelModule
        item = KernelModule('floppy')
        if item.is_running():
            item.stop()
        item.start()
        self.assertTrue(item.is_running())

    def test_stop(self):
        from . import KernelModule
        item = KernelModule('floppy')
        if not item.is_running():
            item.start()
        item.stop()
        self.assertFalse(item.is_running())
        item.start()

    def test_stop__module_in_use(self):
        from . import KernelModule, ServiceFailedToStop
        from logging import getLogger
        debug = getLogger(__name__).debug
        module_in_use = self._get_module_in_use()
        debug("module is use: %s", module_in_use)
        if module_in_use is None:
            raise unittest.SkipTest
        item = KernelModule(module_in_use)
        self.assertRaises(ServiceFailedToStop, item.stop)

class MockInitServiceTestCase(InitServiceTestCase):
    def should_skip(self):
        return False

    def _choose_service(self):
        self.service_name = 'foo'

    @contextmanager
    def mock_exists(self):
        from os.path import sep
        def side_effect(*args, **kwargs):
            path = args[0]
            return '20' in path.split(sep)

        with mock.patch("os.path.exists") as exists:
            exists.side_effect = side_effect
            yield exists

    @contextmanager
    def mock_open(self):
        with mock.patch("__builtin__.open") as _open:
            _open.return_value = mock.MagicMock(file)
            _open.return_value.__enter__.return_value.read.return_value = '20\n'
            yield _open

    def test_get_list_of_running_processes(self):
        from os.path import join, sep
        with nested(mock.patch("os.listdir"),
                    mock.patch("os.readlink"),
                    self.mock_exists()) \
                    as (listdir, readlink, _):
            listdir.return_value = ['10', 'a', '20', 'b', '30', 'c']
            readlink.return_value = join(sep, 'sbin', 'foo')
            InitServiceTestCase.test_get_list_of_running_processes(self)

    def test_is_installed(self):
        with mock.patch("os.path.exists") as exists:
            exists.return_value = True
            InitServiceTestCase.test_is_installed(self)

    def test_is_running(self):
        with nested(mock.patch("os.listdir"),
                    mock.patch("os.path.exists"),
                    self.mock_open()) \
                    as (listdir, exists, _):
            listdir.return_value = ['10', 'a', '20', 'b', '30', 'c']
            exists.return_value = True
            InitServiceTestCase.test_is_running(self)

    def test_start(self):
        with nested(mock.patch("os.listdir"),
                    mock.patch("os.path.exists"),
                    self.mock_open(),
                    self.mock_execute(0)) \
                    as (listdir, exists, _, _):
            listdir.return_value = ['10', 'a', '20', 'b', '30', 'c']
            exists.return_value = True
            InitServiceTestCase.test_start(self)

    def test_stop(self):
        with nested(mock.patch("os.listdir"),
                    mock.patch("os.path.exists"),
                    self.mock_open(),
                    self.mock_execute(0)) \
                    as (listdir, exists, _, _):
            listdir.return_value = ['10', 'a', '20', 'b', '30', 'c']
            exists.return_value = False
            InitServiceTestCase.test_stop(self)
            exists.return_value = True
            listdir.return_value = ['10', 'a', 'b', '30', 'c']
            InitServiceTestCase.test_stop(self)

    def test_restart(self):
        pass

    def test_reload(self):
        pass

class MockKernelModuleTestCase(KernelModuleTestCase):
    def setUp(self):
        KernelModuleTestCase.setUp(self)
        self._prepare_modules_dep_file()

    def tearDown(self):
        from os import close, remove
        fd, filename = self.tempfile
        fd.close()
        remove(filename)
        KernelModuleTestCase.tearDown(self)

    def should_skip(self):
        return False

    def _prepare_modules_dep_file(self):
        from os import fdopen
        from tempfile import mkstemp
        fd, filename = mkstemp()
        fd = fdopen(fd, 'w+b')
        fd.write('/legen/wait/for/it/dary.ko:\n')
        fd.write('/foo/bar/floppy.ko:\n')
        fd.write('/abra/kadabra.ko:\n')
        fd.seek(0)
        self.tempfile = fd, filename

    def _get_module_in_use(self):
        return 'floppy'

    @contextmanager
    def mock_uname(self):
        with mock.patch("platform.uname") as uname:
            uname.return_value = ("Linux", "hostname", "2.6.something", "long text", "x64_64", '')
            yield uname

    @contextmanager
    def mock_open(self):
        fd, _ = self.tempfile
        with mock.patch("__builtin__.open") as _open:
            _open.return_value = mock.MagicMock(fd)
            _open.return_value.__enter__.return_value.read.return_value = fd.read()
            _open.return_value.__enter__.return_value.readlines.return_value = []
            yield _open

    def test_is_installed(self):
        with nested(self.mock_uname(), self.mock_execute(0), self.mock_open()) as (_, _, _open):
            _open.return_value.__enter__.return_value.readlines.return_value = \
                    ["/a/b/c/floppy.ko: a\n/foo/a.ko:"]
            KernelModuleTestCase.test_is_installed(self)
            self.assertEquals(_open.return_value.__enter__.return_value.read.call_count, 1)


    def test_start(self):
        with nested(self.mock_uname(), self.mock_execute(0), self.mock_open()) as (_, _, _open):
            return_values = [['ext3 4321 0 - Live 0x00'],
                             ['floppy 1234 0 - Live 0x00', 'ext3 4321 0 - Live 0x00']]
            return_values.reverse()

            def side_effect(*args, **kwargs):
                return return_values.pop()

            _open.return_value.__enter__.return_value.readlines.side_effect = side_effect
            KernelModuleTestCase.test_start(self)

    def test_stop(self):
        with nested(self.mock_uname(), self.mock_execute(0), self.mock_open()) as (_, _, _open):
            return_values = [['ext3 4321 0 - Live 0x00'],
                             ['floppy 1234 0 - Live 0x00', 'ext3 4321 0 - Live 0x00']]

            def side_effect(*args, **kwargs):
                return return_values.pop()

            _open.return_value.__enter__.return_value.readlines.side_effect = side_effect
            KernelModuleTestCase.test_stop(self)

    def test_is_installed__no_such_module(self):
        with nested(self.mock_uname(), self.mock_execute(0), self.mock_open()):
            KernelModuleTestCase.test_is_installed__no_such_module(self)

    def test_stop__module_in_use(self):
        with nested(self.mock_uname(), self.mock_execute(1), self.mock_open()):
            KernelModuleTestCase.test_stop__module_in_use(self)

class SimpleCompositeServiceTestCase(unittest.TestCase):
    def setUp(self):
        from . import CompositeEntryPoint
        self.concrete = CompositeEntryPoint()
        self.concrete.add_component(MockEntryPoint(), 1)
        self.concrete.add_component(MockEntryPoint(), 5)
        self.concrete.add_component(MockEntryPoint(), 10)

    def test_stop_start(self):
        self.concrete.start()
        self.concrete.stop()

    def test_restart(self):
        self.concrete.restart()

    def test_reload(self):
        self.concrete.reload()

    def test_is_running(self):
        self.concrete.stop()
        self.assertFalse(self.concrete.is_running())
        self.concrete.start()
        self.assertTrue(self.concrete.is_running())

    def test_is_installed(self):
        self.assertTrue(self.concrete.is_installed())
        for key, value in self.concrete.iter_components_by_order():
            value._installed = False
        self.assertFalse(self.concrete.is_installed())

    def test_remove_componet(self):
        for key, value in self.concrete.iter_components_by_order():
            self.concrete.remove_component(value)
        self.concrete.remove_component(EntryPoint())

    def test_add_component(self):
        component = MockEntryPoint()
        self.concrete.add_component(component, 100)
        self.concrete.add_component(component, 100)
        self.assertIn(component, self.concrete.components.values())
        self.concrete.remove_component(component)
        self.assertNotIn(component, self.concrete.components.values())




from infi.exceptools import chain
from infi.instruct import Struct
from infi.instruct import ULInt32, ULInt64

from os.path import exists, sep, join, exists
from ctypes import c_size_t, sizeof

DEFAULT_SOCKET = join(sep, 'var', 'run', 'multipathd.sock')
if not exists(DEFAULT_SOCKET) and exists("/etc/systemd"):
    DEFAULT_SOCKET = "\x00/org/kernel/linux/storage/multipathd"

DEFAULT_TIMEOUT = 60
MAX_SIZE = 2 ** 8
HEADER_SIZE = sizeof(c_size_t)

from logging import getLogger
logger = getLogger(__name__)


try:
    from gevent import socket
except ImportError:
    import socket

class ClientBaseException(Exception):
    pass

from ..errors import ConnectionError, TimeoutExpired

class MessageLength(Struct):
    _fields_ = [(ULInt64 if HEADER_SIZE == 8 else ULInt32)("length"), ]

class BaseConnection(object):
    def __init__(self):
        super(BaseConnection, self).__init__() #pragma: no cover

    def connect(self):
        raise NotImplementedError #pragma: no cover

    def send(self, message):
        raise NotImplementedError #pragma: no cover

    def receive(self, message):
        raise NotImplementedError #pragma: no cover

    def disconnect(self):
        raise NotImplementedError #pragma: no cover

class UnixDomainSocket(BaseConnection):
    def __init__(self, timeout=DEFAULT_TIMEOUT, address=DEFAULT_SOCKET):
        super(UnixDomainSocket, self).__init__() #pragma: no cover
        self._timeout = timeout
        self._address = address
        self._socket = None

    def connect(self):
        try:
            socket_object = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            socket_object.connect(self._address)
            socket_object.settimeout(self._timeout)
        except socket.error, error:
            raise chain(ConnectionError("failed to connect to multipathd socket: {}".format(error.strerror)))
        except Exception, error:
            raise chain(ConnectionError("failed to connect to multipathd socket: {}".format(error)))
        self._socket = socket_object

    def send(self, message):
        try:
            bytes_sent = self._socket.send(message)
        except socket.timeout:
            logger.debug("Caught socket timeout: {!r}".format(self._socket.gettimeout()))
            raise chain(TimeoutExpired("multipathd is not responding"))
        except socket.error, error:
            raise chain(ConnectionError("multipathd socket error: {}".format(error.strerror)))

        if bytes_sent < len(message):
            self.send(message[bytes_sent:])

    def _receive(self, expected_length=MAX_SIZE):
        received_string = ''
        remaining_length = expected_length
        while remaining_length > 0:
            unit_length = MAX_SIZE if remaining_length > MAX_SIZE else remaining_length
            received_string += self._socket.recv(unit_length)
            remaining_length = expected_length - len(received_string)
        return received_string

    def receive(self, expected_length=MAX_SIZE):
        try:
            return self._receive(expected_length)
        except socket.timeout:
            logger.debug("Caught socket timeout: {!r}".format(self._socket.gettimeout()))
            raise chain(TimeoutExpired("multipathd is not responding"))
        except socket.error, error:
            raise chain(ConnectionError("multipathd connection error: {}".format(error.strerror)))

    def disconnect(self):
        if self._socket is not None:
            self._socket.close()
        self._socket = None

    def __repr__(self):
        try:
            msg = "<UnixDomainSocket(timeout={}, address={}, socket={})>"
            return msg.format(self._timeout, self._address, self._socket)
        except:
            return super(UnixDomainSocket, self).__repr__()




from ..connection import BaseConnection
from ..connection import MessageLength
from infi.exceptools import chain

#pylint: disable-msg=E1101

class SimulatorError(Exception):
    pass

class SimulatorConnection(BaseConnection):
    def __init__(self):
        self._connected = False
        self._response = None
        self._expecting_length = False
        self._expected_length = 0
        self._message_to_return = None
        self._sent_message_size = False

    def connect(self):
        self._connected = True
        self._response = None
        self._expecting_length = True
        self._expected_length = 0
        self._message_to_return = None
        self._sent_message_size = False

    def send(self, message):
        if self._expecting_length:
            self._expected_length = MessageLength.create_from_string(message).length
            self._expecting_length = False
        else:
            assert len(message) == self._expected_length
            self._message_to_return = Singleton().handle_incomming_message(message)

    def receive(self, expected_size): #pylint: disable-msg=W0613
        if not self._connected or not self._message_to_return:
            raise SimulatorError
        if not self._sent_message_size:
            self._sent_message_size = True
            instance = MessageLength()
            instance.length = len(self._message_to_return)
            return MessageLength.write_to_string(instance)
        else:
            return self._message_to_return

    def disconnect(self):
        self._connected = False

class Simulator(object):
    def __init__(self, conf_path):
        super(Simulator, self).__init__()
        self._handled_messages = []
        self._conf_path = conf_path
        self._write_sample_config_if_empty()
        self._load_configuration()
        self._load_devices()

    def _load_devices(self):
        from ..model import tests
        from ..model import get_list_of_multipath_devices_from_multipathd_output
        from mock import patch
        import __builtin__
        output = tests.MOCK_OUTPUT[-1]
        maps_topology = output['show multipaths topology']
        paths_table = output['show paths']
        with patch.object(__builtin__, "open"):
            self._devices = get_list_of_multipath_devices_from_multipathd_output(maps_topology, paths_table)

    def _write_sample_config_if_empty(self):
        from os.path import exists, join, sep, dirname
        if exists(self._conf_path) and open(self._conf_path).read() != '':
            return
        from ..config.tests import SAMPLE_FILEPATH
        with open(SAMPLE_FILEPATH) as sample_fd:
            sample_content = sample_fd.read()
        with open(self._conf_path, 'w') as conf_fd:
            conf_fd.write(sample_content)

    def _load_configuration(self):
        from ..config import Configuration
        with open(self._conf_path) as conf_fd:
            self._configuration = Configuration.from_multipathd_conf(conf_fd.read())

    def _path_change_state(self, path_id):
        for device in self._devices:
            for pathgroup in device.path_groups:
                for path in pathgroup.paths:
                    if path.id == path_id:
                        print 'changing state of %s from %s' % (path_id, path.state)
                        path.state = 'failed' if path.state == 'active' else 'active'

    def handle_incomming_message(self, message):
        from ..model import tests
        message = message.strip('\n')
        self._handled_messages.append(message)
        print message
        if message == 'show config':
            return self._configuration.to_multipathd_conf()
        if message == 'reconfigure':
            self._load_configuration()
            return 'ok\n'
        if message in ['show multipaths topology', 'show paths']:
            output = tests.MOCK_OUTPUT[-1]
            return output[message]
        if message.rsplit(' ', 1)[0] == 'fail path':
            path_id = message.rsplit(' ', 1)[1]
            self._path_change_state(path_id)
            return 'ok\n'
        if message.rsplit(' ', 1)[0] == 'reinstate path':
            path_id = message.rsplit(' ', 1)[1]
            self._path_change_state(path_id)
            return 'ok\n'
        if message.rsplit(' ', 1)[0] == '?':
            from ..model.tests import VERSION_OUTPUT
            return VERSION_OUTPUT[0]

#TODO load some multipaths to simualtor
#TODO support fail_path, reinstate_path

class Singleton(object):
    _instance = None

    @classmethod
    def __new__(cls, *args, **kwargs): #pylint: disable-msg=W0613
        if not cls._instance:
            from tempfile import mkstemp
            from os import close
            fd, fpath = mkstemp()
            close(fd)
            cls._instance = Simulator(fpath)
        return cls._instance




import unittest
import mock
from contextlib import contextmanager
from logging import getLogger

loger = getLogger(__name__)

from ctypes import sizeof, c_size_t

#pylint: disable-all

TEST_MESSAGE_TO_SEND = 'reconfigure'
TEST_MESSAGE_SIZE_TO_SEND = '\x0b' + '\x00' * (sizeof(c_size_t) - 1)
TEST_MESSAGE_SIZE_TO_RECEIVE = '\x04' + '\x00' * (sizeof(c_size_t) - 1)
TEST_MESSAGE_TO_RECEIVE = 'ok\n\x00'

class MultipathClientTestCase(unittest.TestCase):
    def setUp(self):
        from platform import system
        from . import MultipathClient
        if system().lower() != 'linux':
            raise unittest.SkipTest
        super(MultipathClientTestCase, self).setUp()
        self.client = MultipathClient()

    def test_rescan(self):
        self.client.rescan()

    def test_is_running(self):
        self.assertTrue(self.client.is_running())

    def test_get_config(self):
        from ..config import Configuration
        config = self.client.get_multipathd_conf()
        self.assertIsInstance(config, Configuration)

    def test_write_config(self):
        config = self.client.get_multipathd_conf()
        self.client.write_to_multipathd_conf(config)

    def test_get_devices(self):
        import __builtin__
        with mock.patch.object(__builtin__, "open"):
            devices = self.client.get_list_of_multipath_devices()

    def test_disable_and_reinstante_paths(self):
        raise unittest.SkipTest
        from time import sleep
        devices = self.client.get_list_of_multipath_devices()
        path_id = devices[0].path_groups[0].paths[0].id
        self.client.fail_path(path_id)
        sleep(1)
        devices = self.client.get_list_of_multipath_devices()
        path = devices[0].path_groups[0].paths[0]
        self.assertEqual(path.id, path_id)
        self.assertEqual(path.state, 'failed')
        self.client.reinstate_path(path.id)
        sleep(1)
        devices = self.client.get_list_of_multipath_devices()
        path = devices[0].path_groups[0].paths[0]
        self.assertEqual(path.id, path_id)
        self.assertEqual(path.state, 'active')

class MutipathClientSimulatorTestCase(MultipathClientTestCase):
    def setUp(self):
        from ..simulator import SimulatorConnection, Singleton
        from . import MultipathClient
        super(MultipathClientTestCase, self).setUp()
        self.client = MultipathClient(SimulatorConnection())
        self.simulator = Singleton()

    def test_rescan(self):
        super(MutipathClientSimulatorTestCase, self).test_rescan()
        self.assertIn('reconfigure', self.simulator._handled_messages)

    def test_get_config(self):
        super(MutipathClientSimulatorTestCase, self).test_get_config()
        self.assertIn('show config', self.simulator._handled_messages)

    def test_write_config(self):
        config = self.client.get_multipathd_conf()
        self.client.write_to_multipathd_conf(config, self.simulator._conf_path)

    def test_disable_and_reinstante_paths(self):
        raise unittest.SkipTest

    def test_get_version(self):
        actual = self.client.get_version()
        expected = [0, 4, 8]
        self.assertEqual(actual, expected)
